// Description: 
// By Sarah Dean
// Email: sdean12@softhome.net
// WWW:   http://www.FreeOTFE.org/
//
// -----------------------------------------------------------------------------
//


#include "FreeOTFECypherImpl.h"
#include "FreeOTFECypherCAST5.h"

#include "FreeOTFEDebug.h"
#include "FreeOTFElib.h"


// Include libtomcrypt CAST5 library...
//#define CBC 1
//#include <mycrypt.h>
//#include <mycrypt_cipher.h>
#include <mycrypt_custom.h>


// =========================================================================
// Cypher driver init function
// devExt - The device extension to be initialized
NTSTATUS
ImpCypherDriverExtDetailsInit(
    IN OUT DEVICE_EXTENSION* devExt
)
{
    NTSTATUS status = STATUS_SUCCESS;
    int idx = -1;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherDriverExtDetailsInit\n"));


    // -- POPULATE DRIVER IDENTIFICATION --
    RtlZeroMemory(devExt->DriverTitle, sizeof(devExt->DriverTitle));
    RtlCopyMemory(
                  devExt->DriverTitle,
                  DRIVER_TITLE,
                  strlen(DRIVER_TITLE)
                 );

    devExt->DriverGUID = DRIVER_GUID;
    devExt->DriverVersionID = DRIVER_CYPHER_VERSION;


    // -- POPULATE cyphers SUPPORTED --
    devExt->CypherDetails = ExAllocatePool(
                                         NonPagedPool,
                                         (sizeof(CYPHER) * CYPHERS_SUPPORTED)
                                        );    

    devExt->CypherCount = CYPHERS_SUPPORTED;


    // -- CAST5 --
    idx++;
    RtlZeroMemory(devExt->CypherDetails[idx].Title, MAX_CYPHER_TITLE);
    RtlCopyMemory(
                  devExt->CypherDetails[idx].Title,
                  DRIVER_CIPHER_TITLE_CAST5,
                  strlen(DRIVER_CIPHER_TITLE_CAST5)
                 );

    devExt->CypherDetails[idx].Mode       = CYPHER_MODE_CBC;
    devExt->CypherDetails[idx].KeySize    = 128;
    devExt->CypherDetails[idx].BlockSize  = (cast5_desc.block_length * 8);
    devExt->CypherDetails[idx].VersionID  = DRIVER_CYPHER_IMPL_VERSION;
    devExt->CypherDetails[idx].CypherGUID = CIPHER_GUID_CAST5;


    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherDriverExtDetailsInit\n"));

    return status;
}


// =========================================================================
// Cypher driver cleardown function
// devExt - The device extension to be cleared down
NTSTATUS
ImpCypherDriverExtDetailsCleardown(    
    IN OUT DEVICE_EXTENSION* devExt
)
{
    NTSTATUS status = STATUS_SUCCESS;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherDriverExtDetailsCleardown\n"));

    if (devExt->CypherDetails != NULL)
        {
        ExFreePool(devExt->CypherDetails);
        }

    devExt->CypherDetails = NULL;
    devExt->CypherCount = 0;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherDriverExtDetailsCleardown\n"));

    return status;
}


// =========================================================================
// Encryption function
// Note: CyphertextLength must be set to the size of the CyphertextData buffer on
//       entry; on exit, this will be set to the size of the buffer used.
NTSTATUS
ImpCypherEncryptData(
    IN      GUID* CypherGUID,
    IN      int KeyLength,  // In bits
    IN      char *Key,
    IN      char *KeyASCII,  // ASCII representation of "Key"
    IN      int IVLength,  // In bits
    IN      char *IV,
    IN      int PlaintextLength,  // In bytes
    IN      char *PlaintextData,
    OUT     char *CyphertextData
)
{
    NTSTATUS status = STATUS_SUCCESS;
    // libtomcrypt can't handle NULL IVs in CBC mode - it ASSERTs that IV != NULL
    char ltcNullIV[FREEOTFE_MAX_CYPHER_BLOCKSIZE];
    int cipher;
    unsigned int i;
    symmetric_CBC *cbc;
    int errnum;
    unsigned char* inBufPtr;
    unsigned char* outBufPtr;
    unsigned int x;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherEncryptData\n"));

    if (!(
		  (IsEqualGUID(&CIPHER_GUID_CAST5, CypherGUID))
		 ))
		{
		DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Unsupported cipher GUID passed in.\n"));
        status = STATUS_INVALID_PARAMETER;
        }

    // libtomcrypt can't handle NULL IVs in CBC mode - it ASSERTs that IV != NULL
    if ( (IVLength == 0) || (IV == NULL) )
        {
        RtlZeroMemory(&ltcNullIV, sizeof(ltcNullIV));
        IV = (char*)&ltcNullIV;
        }

    cbc = ExAllocatePool(
                         NonPagedPool,
                         sizeof(symmetric_CBC)
                        );    
    RtlZeroMemory(cbc, sizeof(symmetric_CBC));

	if NT_SUCCESS(status) 
		{
		status = InitLTCCypher(&cipher);
		}

	if NT_SUCCESS(status)
		{
	    // Start a CBC session
        if ((errnum = cbc_start(cipher, IV, Key, (KeyLength/8), 0, cbc)) != CRYPT_OK)
            {
			status = STATUS_UNSUCCESSFUL;
			DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Unable to start CBC session (errnum: %d)\n", errnum));
            }
		}

    if NT_SUCCESS(status)
        {
        x = (unsigned int)(int)(PlaintextLength / cast5_desc.block_length);
		inBufPtr = PlaintextData;
		outBufPtr = CyphertextData;
        for (i = 0; i < x; i++)
            {
            if ((errnum = cbc_encrypt(inBufPtr, outBufPtr, cbc)) != CRYPT_OK)
                {
                DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Unable to encrypt/decrypt block (errnum: %d)\n", errnum));
                status = STATUS_UNSUCCESSFUL;
                } 
                
            inBufPtr  += cast5_desc.block_length;
            outBufPtr += cast5_desc.block_length;
            }
		}

    SecZeroMemory(cbc, sizeof(symmetric_CBC));
    ExFreePool(cbc);


    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherEncryptData\n"));

    return status;
}


// =========================================================================
// Decryption function
// Note: PlaintextLength must be set to the size of the PlaintextData buffer on
//       entry; on exit, this will be set to the size of the buffer used.
NTSTATUS
ImpCypherDecryptData(
    IN      GUID* CypherGUID,
    IN      int KeyLength,  // In bits
    IN      char *Key,
    IN      char *KeyASCII,  // ASCII representation of "Key"
    IN      int IVLength,  // In bits
    IN      char *IV,
    IN      int CyphertextLength,  // In bytes
    IN      char *CyphertextData,
    OUT     char *PlaintextData
)
{
    NTSTATUS status = STATUS_SUCCESS;
    // libtomcrypt can't handle NULL IVs in CBC mode - it ASSERTs that IV != NULL
    char ltcNullIV[FREEOTFE_MAX_CYPHER_BLOCKSIZE];
    int cipher;
    unsigned int i;
    symmetric_CBC *cbc;
    int errnum;
    unsigned char* inBufPtr;
    unsigned char* outBufPtr;
    unsigned int x;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherDecryptData\n"));

	if (!(
		  (IsEqualGUID(&CIPHER_GUID_CAST5, CypherGUID))
		 ))
		{
		DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Unsupported cipher GUID passed in.\n"));
        status = STATUS_INVALID_PARAMETER;
        }

    // libtomcrypt can't handle NULL IVs in CBC mode - it ASSERTs that IV != NULL
    if ( (IVLength == 0) || (IV == NULL) )
        {
        RtlZeroMemory(&ltcNullIV, sizeof(ltcNullIV));
        IV = (char*)&ltcNullIV;
        }

    cbc = ExAllocatePool(
                         NonPagedPool,
                         sizeof(symmetric_CBC)
                        );    
    RtlZeroMemory(cbc, sizeof(symmetric_CBC));

	if NT_SUCCESS(status) 
		{
		status = InitLTCCypher(&cipher);
		}

	if NT_SUCCESS(status)
		{
	    // Start a CBC session
        if ((errnum = cbc_start(cipher, IV, Key, (KeyLength/8), 0, cbc)) != CRYPT_OK)
            {
			status = STATUS_UNSUCCESSFUL;
			DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Unable to start CBC session (errnum: %d)\n", errnum));
            }
		}

    if NT_SUCCESS(status)
        {
        x = (unsigned int)(int)(CyphertextLength / cast5_desc.block_length);
		inBufPtr = CyphertextData;
		outBufPtr = PlaintextData;
        for (i = 0; i < x; i++)
            {
            if ((errnum = cbc_decrypt(inBufPtr, outBufPtr, cbc)) != CRYPT_OK)
                {
                DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Unable to encrypt/decrypt block (errnum: %d)\n", errnum));
                status = STATUS_UNSUCCESSFUL;
                } 
                
            inBufPtr  += cast5_desc.block_length;
            outBufPtr += cast5_desc.block_length;
            }
        }


    SecZeroMemory(cbc, sizeof(symmetric_CBC));
    ExFreePool(cbc);

    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherDecryptData\n"));

    return status;
}


// =========================================================================
// Initialize libtomcrypt cypher
NTSTATUS
InitLTCCypher(
    OUT  int *cipher
)
{
	NTSTATUS status = STATUS_CRYPTO_SYSTEM_INVALID;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("InitLTCCypher\n"));

    // Initialize cipher
    *cipher = register_cipher(&cast5_desc);
    if (*cipher == -1)
        {
	    DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Could not register cipher\n"));
        }
    else
        {    
        *cipher = find_cipher("cast5");
        if (*cipher == -1)
            {
      	    DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Could not find cipher\n"));
            }
		else
			{
			status = STATUS_SUCCESS;
			}
        }

    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("InitLTCCypher\n"));

	return status;
}


// =========================================================================
// =========================================================================

