// Description: FreeOTFE Device Driver API
// By Sarah Dean
// Email: sdean12@softhome.net
// WWW:   http://www.FreeOTFE.org/
//
// -----------------------------------------------------------------------------
//


#ifndef _FreeOTFEAPI_H
#define _FreeOTFEAPI_H   1


// Required for definition of ULONG
#include <windef.h>

// Required for definition of STORAGE_MEDIA_TYPE
#include <ntddstor.h>


// Following is required for the "_T" macro
/// #include <tchar.h>

// #define _T(x) L ## x

// =========================================================================
// Const definitions


// Length in characters (not bytes)
// This is the max number of chars in:
//   *) A FreeOTFE device name
//   *) A volume's filename
// There's no *real* reason to use, this - and this does create a limitation (YUK!)
// The only reason this is used is so that creating and populating the DIOC buffers is
// slightly easier (the stuct lengths are more deterministic with this const in place)
// Things may be rewritten slightly in the future to remove this const
// xxxop - remove this const; dynamically alloc DIOC buffer which uses it
#define FREEOTFE_MAX_FILENAME_LENGTH   1024

#define DEVICE_BASE_NAME              L"\\FreeOTFE"
#define DEVICE_FREEOTFE_ROOT          L"\\Device" DEVICE_BASE_NAME
#define DEVICE_SYMLINK_FREEOTFE_ROOT  L"\\DosDevices"

// Main device
#define DEVICE_MAIN_NAME              DEVICE_FREEOTFE_ROOT DEVICE_BASE_NAME
#define DEVICE_SYMLINK_MAIN_NAME      DEVICE_SYMLINK_FREEOTFE_ROOT DEVICE_BASE_NAME

// Disk devices
#define DEVICE_DISK_DIR_NAME          DEVICE_FREEOTFE_ROOT L"\\Disks"
#define DEVICE_DISK_PREFIX            DEVICE_DISK_DIR_NAME L"\\Disk"
// The following is just padding to represent the number tacked onto the
// end of DEVICE_DISK_PREFIX
#define DEVICE_DISK_NUMBER            L"zzzzz"


// The size of the critical data block
#define CRITICAL_DATA_LENGTH  4096  // In *bits*



// VolumeFlags bits:
#define VOL_FLAGS_USE_SECTOR_ID_AS_IV        1  // Bit 0
#define VOL_FLAGS_SECTOR_ID_ZERO_VOLSTART    2  // Bit 1
// Bit 2 unused
#define VOL_FLAGS_HASH_SECTOR_ID_BEFORE_USE  8  // Bit 3

  


#define IOCTL_FREEOTFE_VERSION         CTL_CODE(                   \
                                                FILE_DEVICE_DISK,  \
                                                0x800,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )

                                               
#define IOCTL_FREEOTFE_CREATE          CTL_CODE(                   \
                                                FILE_DEVICE_DISK,  \
                                                0x805,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )
                                               
#define IOCTL_FREEOTFE_SET_RAW         CTL_CODE(                   \
                                                FILE_DEVICE_DISK,  \
                                                0x809,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )
                                               
#define IOCTL_FREEOTFE_GET_RAW         CTL_CODE(                   \
                                                FILE_DEVICE_DISK,  \
                                                0x80A,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )
                                               
#define IOCTL_FREEOTFE_MOUNT           CTL_CODE(                   \
                                                FILE_DEVICE_DISK,  \
                                                0x801,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )

#define IOCTL_FREEOTFE_DISMOUNT        CTL_CODE(                   \
                                                FILE_DEVICE_DISK,  \
                                                0x802,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )

#define IOCTL_FREEOTFE_DESTROY         CTL_CODE(                   \
                                                FILE_DEVICE_DISK,  \
                                                0x806,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )

// This returns the number of *disk* devices which exist
#define IOCTL_FREEOTFE_GET_DISK_DEVICE_COUNT CTL_CODE(             \
                                                FILE_DEVICE_DISK,  \
                                                0x807,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )

// This returns the kernel mode devicenames of all *disk* devices which exist
#define IOCTL_FREEOTFE_GET_DISK_DEVICE_LIST CTL_CODE(              \
                                                FILE_DEVICE_DISK,  \
                                                0x808,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )

// This returns status information on the named *disk* device
#define IOCTL_FREEOTFE_GET_DISK_DEVICE_STATUS  CTL_CODE(           \
                                                FILE_DEVICE_DISK,  \
                                                0x803,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )

// This returns metadata on the named *disk* device, as passed in when it was mounted
#define IOCTL_FREEOTFE_GET_DISK_DEVICE_METADATA  CTL_CODE(         \
                                                FILE_DEVICE_DISK,  \
                                                0x804,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )



// Derive a key
#define IOCTL_FREEOTFE_DERIVE_KEY      CTL_CODE(                   \
                                                FILE_DEVICE_UNKNOWN,  \
                                                0x810,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )

// Generate a MAC
#define IOCTL_FREEOTFE_GENERATE_MAC    CTL_CODE(                   \
                                                FILE_DEVICE_UNKNOWN,  \
                                                0x811,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )

// Encrypt a block
#define IOCTL_FREEOTFE_ENCRYPT_BLOCK    CTL_CODE(                   \
                                                FILE_DEVICE_UNKNOWN,  \
                                                0x812,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )

// Decrypt a block
#define IOCTL_FREEOTFE_DECRYPT_BLOCK    CTL_CODE(                   \
                                                FILE_DEVICE_UNKNOWN,  \
                                                0x813,             \
                                                METHOD_BUFFERED,   \
                                                FILE_READ_DATA     \
                                               )


// =========================================================================
// Type definitions

// Mount source
typedef enum _MOUNT_SOURCE {
    MNTSRC_FILE             =    1,
    MNTSRC_PARTITION        =    2,
    MNTSRC_UNKNOWN          = 9999
} MOUNT_SOURCE, *PMOUNT_SOURCE;


// IV generation method
typedef enum _SECTOR_IV_GEN_METHOD {
    SCTRIVGEN_NONE                 =   0,
    SCTRIVGEN_32BIT_SECTOR_ID      =   1,
    SCTRIVGEN_64BIT_SECTOR_ID      =   2,
    SCTRIVGEN_HASH_32BIT_SECTOR_ID =   3,
    SCTRIVGEN_HASH_64BIT_SECTOR_ID =   4,
    SCTRIVGEN_ESSIV                =   5,
    SCTRIVGEN_UNKNOWN              = 255
} SECTOR_IV_GEN_METHOD, *PSECTOR_IV_GEN_METHOD;


// Key derivation algorithm to use
typedef enum _KDF_ALGORITHM {
    KDF_HASH     =    1,
    KDF_PBKDF2   =    2,
    KDF_UNKNOWN  = 9999
} KDF_ALGORITHM, *PKDF_ALGORITHM;

// MAC algorithm to use
typedef enum _MAC_ALGORITHM {
    MAC_HASH     =    1,
    MAC_HMAC     =    2,
    MAC_UNKNOWN  = 9999
} MAC_ALGORITHM, *PMAC_ALGORITHM;


typedef struct _DIOC_VERSION {
    ULONG VersionID;    
} DIOC_VERSION, *PDIOC_VERSION;


typedef struct _DIOC_DISK_DEVICE_CREATE {
    ULONG  DeviceType;  // Set to one of the system defined FILE_DEVICE_XXX
                        // constants
} DIOC_DISK_DEVICE_CREATE, *PDIOC_DISK_DEVICE_CREATE;


// This struct returns the device kernel mode name which can then be passed
// in a "DefineDosDevice" call in user mode
typedef struct _DIOC_DEVICE_NAME {
    CHAR DeviceName[FREEOTFE_MAX_FILENAME_LENGTH];
} DIOC_DEVICE_NAME, *PDIOC_DEVICE_NAME;


// This struct returns a list of device kernel mode names
typedef struct _DIOC_DEVICE_NAME_LIST {
    ULONG DeviceCount;
    CHAR DeviceName[1][FREEOTFE_MAX_FILENAME_LENGTH];  // Variable number of device names
} DIOC_DEVICE_NAME_LIST, *PDIOC_DEVICE_NAME_LIST;


typedef struct _DIOC_MOUNT {
    CHAR DiskDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];

    CHAR Filename[FREEOTFE_MAX_FILENAME_LENGTH];    

    // Start and end offsets of encrypted partition within the file
    // If DataEnd is zero, it will be set to the last byte in the file
    LARGE_INTEGER DataStart;
    LARGE_INTEGER DataEnd;

    CHAR IVHashDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];    
    GUID IVHashGUID;
    CHAR IVCypherDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];    
    GUID IVCypherGUID;
    
    CHAR MainCypherDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];    
    GUID MainCypherGUID;

    BOOLEAN ReadOnly;
    MOUNT_SOURCE MountSource;  // Identifies what the encrypted data is stored on
    STORAGE_MEDIA_TYPE StorageMediaType;  // Identifies what the virtual device should emulate

    unsigned int VolumeFlags;
    SECTOR_IV_GEN_METHOD SectorIVGenMethod;

    // NOTE: THE ORDER OF THE FOLLOWING ITEMS IS SIGNIFICANT
    //       We have the lengths first, *then* the data as this makes checking the size of the
    //       struct easier, since these are both known-width members and can therefore
    //       be accessed directly using "variable->VolumeIVLength".
    //       If the order was: "MasterKey", "VolumeIVLength", "VolumeIV", then we could not access
    //       "VolumeIVLength" directly, and would have to take "MasterKeyLength" into account before
    //       accessing it.
    ULONG MasterKeyLength;  // In bits
    ULONG VolumeIVLength;  // In bits
    ULONG MetaDataLength;  // In bytes
    CHAR MasterKey[1];    
    CHAR VolumeIV[1];    
    CHAR MetaData[1];
} DIOC_MOUNT, *PDIOC_MOUNT;


typedef struct _DIOC_DISMOUNT {
    CHAR DiskDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];
    BOOLEAN Emergency;
} DIOC_DISMOUNT, *PDIOC_DISMOUNT;


typedef struct _DIOC_DISK_DEVICE_COUNT {
    ULONG Count;    
} DIOC_DISK_DEVICE_COUNT, *PDIOC_DISK_DEVICE_COUNT;



typedef struct _DIOC_DISK_DEVICE_STATUS {
    CHAR DeviceName[FREEOTFE_MAX_FILENAME_LENGTH];
    CHAR Filename[FREEOTFE_MAX_FILENAME_LENGTH];    
    BOOLEAN Mounted;
    
    CHAR IVHashDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];    
    GUID IVHashGUID;
    CHAR IVCypherDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];    
    GUID IVCypherGUID;
    
    CHAR MainCypherDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];    
    GUID MainCypherGUID;

    BOOLEAN DismountPending;
    BOOLEAN ReadOnly;

    unsigned int VolumeFlags;
    SECTOR_IV_GEN_METHOD SectorIVGenMethod;

    ULONG MetaDataLength;  // In bytes
} DIOC_DISK_DEVICE_STATUS, *PDIOC_DISK_DEVICE_STATUS;


typedef struct _DIOC_DISK_DEVICE_METADATA {
    ULONG MetaDataLength;  // In bytes
    CHAR MetaData[1];
} DIOC_DISK_DEVICE_METADATA, *PDIOC_DISK_DEVICE_METADATA;


typedef struct _DIOC_SET_RAW_DATA {
    CHAR Filename[FREEOTFE_MAX_FILENAME_LENGTH];
    LARGE_INTEGER Offset;
    ULONG DataLength;
    BYTE Data[1];  // Variable length
} DIOC_SET_RAW_DATA, *PDIOC_SET_RAW_DATA;


typedef struct _DIOC_GET_RAW_DATA_IN {
    CHAR Filename[FREEOTFE_MAX_FILENAME_LENGTH];
    LARGE_INTEGER Offset;
    ULONG DataLength;
} DIOC_GET_RAW_DATA_IN, *PDIOC_GET_RAW_DATA_IN;

typedef struct _DIOC_GET_RAW_DATA_OUT {
    BYTE Data[1];  // Variable length
} DIOC_GET_RAW_DATA_OUT, *PDIOC_GET_RAW_DATA_OUT;


typedef struct _DIOC_DERIVE_KEY_IN {
    KDF_ALGORITHM KDFAlgorithm;

    CHAR HashDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];    
    GUID HashGUID;
    
    CHAR CypherDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];    
    GUID CypherGUID;

    unsigned int Iterations;
    unsigned int LengthWanted;  // In bits - the length of the output key wanted

    // NOTE: THE ORDER OF THE FOLLOWING ITEMS IS SIGNIFICANT
    //       We have the lengths first, *then* the data as this makes checking the size of the
    //       struct easier, since these are both known-width members and can therefore
    //       be accessed directly using "variable->SaltLength".
    //       If the order was: "PasswordLength", "Password", "SaltLength", "Salt", then we could not access
    //       "SaltLength" directly, and would have to take "PasswordLength" into account before
    //       accessing it.
    unsigned int PasswordLength;  // In bits
    unsigned int SaltLength;  // In bits
    unsigned char Password[1];  // Variable length
    unsigned char Salt[1];  // Variable length
} DIOC_DERIVE_KEY_IN, *PDIOC_DERIVE_KEY_IN;

typedef struct _DIOC_DERIVE_KEY_OUT {
    unsigned int DerivedKeyLength;  // In bits
    unsigned char DerivedKey[1];  // Variable length
} DIOC_DERIVE_KEY_OUT, *PDIOC_DERIVE_KEY_OUT;


typedef struct _DIOC_GENERATE_MAC_IN {
    MAC_ALGORITHM MACAlgorithm;

    CHAR HashDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];    
    GUID HashGUID;
    
    CHAR CypherDeviceName[FREEOTFE_MAX_FILENAME_LENGTH];    
    GUID CypherGUID;

    int LengthWanted;

    // NOTE: THE ORDER OF THE FOLLOWING ITEMS IS SIGNIFICANT
    //       We have the lengths first, *then* the data as this makes checking the size of the
    //       struct easier, since these are both known-width members and can therefore
    //       be accessed directly using "variable->SaltLength".
    //       If the order was: "KeyLength", "Key", "DataLength", "Data", then we could not access
    //       "DataLength" directly, and would have to take "KeyLength" into account before
    //       accessing it.
    unsigned int KeyLength;  // In bits
    unsigned int DataLength;  // In bits
    unsigned char Key[1];  // Variable length
    unsigned char Data[1];  // Variable length
} DIOC_GENERATE_MAC_IN, *PDIOC_GENERATE_MAC_IN;

typedef struct _DIOC_GENERATE_MAC_OUT {
    unsigned int MACLength;  // In bits
    unsigned char MAC[1];  // Variable length
} DIOC_GENERATE_MAC_OUT, *PDIOC_GENERATE_MAC_OUT;


// =========================================================================
// =========================================================================

#endif

